/*
 * Copyright (c) 2018 Nordic Semiconductor ASA
 *
 * SPDX-License-Identifier: LicenseRef-Nordic-5-Clause
 */

/** @file
 *  @brief Nordic UART Service Client sample
 */

#include <errno.h>
#include <zephyr/kernel.h>
#include <zephyr/device.h>
#include <zephyr/devicetree.h>
#include <zephyr/sys/byteorder.h>
#include <zephyr/sys/printk.h>

#include <zephyr/bluetooth/bluetooth.h>
#include <zephyr/bluetooth/hci.h>
#include <zephyr/bluetooth/conn.h>
#include <zephyr/bluetooth/uuid.h>
#include <zephyr/bluetooth/gatt.h>

#include <bluetooth/services/nus.h>
#include <bluetooth/services/nus_client.h>
#include <bluetooth/gatt_dm.h>
#include <bluetooth/scan.h>

#include <sdc_hci_vs.h>

/* BLE internal header, use with caution. */
#include <../bluetooth/host/conn_internal.h>

#include <zephyr/settings/settings.h>

#include <zephyr/drivers/uart.h>

#include <zephyr/logging/log.h>

#define LOG_MODULE_NAME central_uart
LOG_MODULE_REGISTER(LOG_MODULE_NAME);

void mtu_exchange_callback(struct bt_conn *conn, uint8_t err,
			   struct bt_gatt_exchange_params *params);

struct bt_gatt_exchange_params exchange_params = {.func = mtu_exchange_callback};

static struct bt_conn *default_conn;
static struct bt_nus_client nus_client;

atomic_t step = 0;
K_SEM_DEFINE(data_received_sem, 0, 1);
static uint8_t on_received(struct bt_conn *conn, struct bt_gatt_subscribe_params *params,
			   const void *data, uint16_t length)
{
	if (NULL != data && length > 0) {
		atomic_set(&step, ((uint8_t *)data)[0] + 1);
		k_sem_give(&data_received_sem);
	}
	LOG_INF("Received data, step is now %ld", atomic_get(&step));
	return BT_GATT_ITER_CONTINUE;
}

static void nus_subscribe(struct k_work *item)
{
	int err;
	static struct bt_gatt_subscribe_params sub_params = {.ccc_handle = 0x13,
							     .value_handle = 0x12,
							     .value = BT_GATT_CCC_NOTIFY,
							     .notify = on_received};

	if (!default_conn) {
		return;
	}

	sub_params.value = BT_GATT_CCC_NOTIFY;

	err = bt_gatt_subscribe(default_conn, &sub_params);
	if (err) {
		LOG_ERR("bt_gatt_subscribe() failed. (err: %d)", err);
		bt_conn_disconnect(default_conn, BT_HCI_ERR_REMOTE_USER_TERM_CONN);
	}
}

static K_WORK_DEFINE(nus_subscribe_work, nus_subscribe);

static void ble_data_sent(struct bt_nus_client *nus, uint8_t err, const uint8_t *const data,
			  uint16_t len)
{
	ARG_UNUSED(nus);

	if (err) {
		LOG_WRN("ATT error code: 0x%02X", err);
	}
}

static uint8_t ble_data_received(struct bt_nus_client *nus, const uint8_t *data, uint16_t len)
{
	ARG_UNUSED(nus);

	LOG_INF("NUS received");

	return BT_GATT_ITER_CONTINUE;
}

static void exchange_func(struct bt_conn *conn, uint8_t err, struct bt_gatt_exchange_params *params)
{
	if (!err) {
		LOG_INF("MTU exchange done");
		k_work_submit(&nus_subscribe_work);
	} else {
		LOG_WRN("MTU exchange failed (err %" PRIu8 ")", err);
		bt_conn_disconnect(conn, BT_HCI_ERR_REMOTE_USER_TERM_CONN);
	}
}

static void connected(struct bt_conn *conn, uint8_t conn_err)
{
	char addr[BT_ADDR_LE_STR_LEN];
	int err;

	bt_addr_le_to_str(bt_conn_get_dst(conn), addr, sizeof(addr));

	if (conn_err) {
		LOG_INF("Failed to connect to %s (%d)", addr, conn_err);

		if (default_conn == conn) {
			bt_conn_unref(default_conn);
			default_conn = NULL;
			k_sem_reset(&data_received_sem);
		}
	} else {
		LOG_INF("Connected: %s", addr);

		static struct bt_gatt_exchange_params exchange_params;

		exchange_params.func = exchange_func;
		err = bt_gatt_exchange_mtu(conn, &exchange_params);
		if (err == -EALREADY) {
			k_work_submit(&nus_subscribe_work);
		} else if (err) {
			LOG_WRN("MTU exchange failed (err %d)", err);

			bt_conn_disconnect(conn, BT_HCI_ERR_REMOTE_USER_TERM_CONN);
		}
	}

	err = bt_scan_start(BT_SCAN_TYPE_SCAN_ACTIVE);
	if (err) {
		LOG_ERR("Scanning failed to start (err %d)", err);
	}
}

static void disconnected(struct bt_conn *conn, uint8_t reason)
{
	char addr[BT_ADDR_LE_STR_LEN];
	int err;

	bt_addr_le_to_str(bt_conn_get_dst(conn), addr, sizeof(addr));

	LOG_INF("Disconnected: %s (reason %u)", addr, reason);

	if (default_conn != conn) {
		return;
	}

	bt_conn_unref(default_conn);
	default_conn = NULL;
	k_sem_reset(&data_received_sem);

	err = bt_scan_start(BT_SCAN_TYPE_SCAN_ACTIVE);
	if (err) {
		LOG_ERR("Scanning failed to start (err %d)", err);
	}
}

BT_CONN_CB_DEFINE(conn_callbacks) = {
	.connected = connected,
	.disconnected = disconnected,
};

#define CONN_LATENCY 0U
/* Arbitrary selected timeout value */
#define CONN_TIMEOUT 400U
static const struct bt_le_conn_param conn_params = BT_LE_CONN_PARAM_INIT(
	BT_GAP_INIT_CONN_INT_MIN, BT_GAP_INIT_CONN_INT_MAX, CONN_LATENCY, CONN_TIMEOUT);
static void scan_filter_match(struct bt_scan_device_info *device_info,
			      struct bt_scan_filter_match *filter_match, bool connectable)
{
	int err;
	char addr[BT_ADDR_LE_STR_LEN];

	bt_addr_le_to_str(device_info->recv_info->addr, addr, sizeof(addr));

	LOG_INF("Filters matched. Address: %s connectable: %d", addr, connectable);

	err = bt_scan_stop();
	if (err != 0) {
		LOG_ERR("bt_scan_stop failed (err %d)", err);
	}

	struct bt_conn_le_create_param conn_param = BT_CONN_LE_CREATE_PARAM_INIT(
		BT_CONN_LE_OPT_NONE, BT_GAP_SCAN_FAST_INTERVAL, BT_GAP_SCAN_FAST_INTERVAL);
	err = bt_conn_le_create(device_info->recv_info->addr, &conn_param, &conn_params,
				&default_conn);
	if (err != 0) {
		LOG_ERR("bt_conn_le_create failed (err %d)", err);
		bt_scan_start(BT_SCAN_TYPE_SCAN_ACTIVE);
	}
}

static void scan_connecting_error(struct bt_scan_device_info *device_info)
{
	LOG_WRN("Connecting failed");
}

static void scan_connecting(struct bt_scan_device_info *device_info, struct bt_conn *conn)
{
	// default_conn = bt_conn_ref(conn);
}

static int nus_client_init(void)
{
	int err;
	struct bt_nus_client_init_param init = {.cb = {
							.received = ble_data_received,
							.sent = ble_data_sent,
						}};

	err = bt_nus_client_init(&nus_client, &init);
	if (err) {
		LOG_ERR("NUS Client initialization failed (err %d)", err);
		return err;
	}

	LOG_INF("NUS Client module initialized");
	return err;
}

BT_SCAN_CB_INIT(scan_cb, scan_filter_match, NULL, scan_connecting_error, scan_connecting);

static int scan_init(void)
{
	int err;
	struct bt_scan_init_param scan_init = {
		.connect_if_match = 0,
	};

	bt_scan_init(&scan_init);
	bt_scan_cb_register(&scan_cb);

	err = bt_scan_filter_add(BT_SCAN_FILTER_TYPE_NAME, "IPC_TEST");
	if (err) {
		LOG_ERR("Scanning name filter cannot be set (err %d)", err);
		return err;
	}

	err = bt_scan_filter_enable(BT_SCAN_NAME_FILTER, false);
	if (err) {
		LOG_ERR("Filters cannot be turned on (err %d)", err);
		return err;
	}

	// err = bt_scan_filter_add(BT_SCAN_FILTER_TYPE_UUID, BT_UUID_NUS_SERVICE);
	// if (err) {
	// 	LOG_ERR("Scanning filters cannot be set (err %d)", err);
	// 	return err;
	// }

	// err = bt_scan_filter_enable(BT_SCAN_UUID_FILTER, false);
	// if (err) {
	// 	LOG_ERR("Filters cannot be turned on (err %d)", err);
	// 	return err;
	// }

	LOG_INF("Scan module initialized");
	return err;
}

bool on_vendor_specific_event(struct net_buf_simple *buf)
{
	uint8_t *subevent_code;
	sdc_hci_subevent_vs_qos_conn_event_report_t *evt_conn;
	bool ret = true;
	bool evt_found = false;

	subevent_code = (uint8_t *)net_buf_simple_pull_mem(buf, sizeof(*subevent_code));

	switch (*subevent_code) {
	case SDC_HCI_SUBEVENT_VS_QOS_CONN_EVENT_REPORT:
		evt_conn = (sdc_hci_subevent_vs_qos_conn_event_report_t *)buf->data;
		evt_found = true;
		break;
	default:
		ret = false;
		break;
	}

	if (!evt_found) {
		return ret;
	}

	LOG_DBG("Conn event report: "
		"handle %d, "
		"event_counter %d, "
		"channel %d, "
		"crc_ok %d, "
		"crc_err %d, "
		"nak_count %d, "
		"rx_timeout %s",
		evt_conn->conn_handle, evt_conn->event_counter, evt_conn->channel_index,
		evt_conn->crc_ok_count, evt_conn->crc_error_count, evt_conn->nak_count,
		((evt_conn->rx_timeout) ? "timedout" : "no timeout"));

	return ret;
}

int conn_event_report_set_enabled(bool enable)
{
	int err;
	struct net_buf *buf;

	err = bt_hci_register_vnd_evt_cb(on_vendor_specific_event);
	if (err) {
		LOG_ERR("Failed to register HCI VS callback");
		return -1;
	}

	sdc_hci_cmd_vs_qos_conn_event_report_enable_t *cmd_enable_conn;

	buf = bt_hci_cmd_create(SDC_HCI_OPCODE_CMD_VS_QOS_CONN_EVENT_REPORT_ENABLE,
				sizeof(*cmd_enable_conn));
	if (!buf) {
		LOG_ERR("Failed to enable HCI VS QoS");
		return -1;
	}

	cmd_enable_conn = (sdc_hci_cmd_vs_qos_conn_event_report_enable_t *)net_buf_add(
		buf, sizeof(*cmd_enable_conn));
	cmd_enable_conn->enable = enable;

	err = bt_hci_cmd_send_sync(SDC_HCI_OPCODE_CMD_VS_QOS_CONN_EVENT_REPORT_ENABLE, buf, NULL);
	if (err) {
		LOG_ERR("Failed to enable HCI VS QoS");
		return -1;
	}

	return 0;
}

void command_sent(struct bt_conn *conn, void *user_data)
{
	ARG_UNUSED(user_data);

	__ASM("nop");
}

#define READ_RSSI_TEST 0
#if READ_RSSI_TEST
void read_conn_rssi(uint16_t handle, int8_t *rssi)
{
	struct net_buf *buf, *rsp = NULL;
	struct bt_hci_cp_read_rssi *cp;
	struct bt_hci_rp_read_rssi *rp;

	int err;

	buf = bt_hci_cmd_create(BT_HCI_OP_READ_RSSI, sizeof(*cp));
	if (!buf) {
		LOG_INF("Unable to allocate command buffer");
		return;
	}

	cp = net_buf_add(buf, sizeof(*cp));
	cp->handle = sys_cpu_to_le16(handle);

	err = bt_hci_cmd_send_sync(BT_HCI_OP_READ_RSSI, buf, &rsp);
	if (err) {
		uint8_t reason = rsp ? ((struct bt_hci_rp_read_rssi *)rsp->data)->status : 0;
		LOG_INF("Read RSSI err: %d reason 0x%02x", err, reason);
		return;
	}

	rp = (void *)rsp->data;
	*rssi = rp->rssi;

	net_buf_unref(rsp);
}
#endif // READ_RSSI_TEST

int main(void)
{
	int err;
	uint32_t pkt_cnt = 0;
	uint8_t data[1000];
#if READ_RSSI_TEST
	int8_t rssi;
#endif

	memset(data, 0x00, sizeof(data));

	err = bt_enable(NULL);
	if (err) {
		LOG_ERR("Bluetooth init failed (err %d)", err);
		return 0;
	}
	LOG_INF("Bluetooth initialized");

	if (IS_ENABLED(CONFIG_SETTINGS)) {
		settings_load();
	}

	err = scan_init();
	if (err != 0) {
		LOG_ERR("scan_init failed (err %d)", err);
		return 0;
	}

	err = conn_event_report_set_enabled(true);
	if (err != 0) {
		LOG_ERR("conn_event_report_set_enabled failed (err %d)", err);
		return 0;
	}

	err = nus_client_init();
	if (err != 0) {
		LOG_ERR("nus_client_init failed (err %d)", err);
		return 0;
	}

	printk("Starting Bluetooth Central UART example\n");

	err = bt_scan_start(BT_SCAN_TYPE_SCAN_ACTIVE);
	if (err) {
		LOG_ERR("Scanning failed to start (err %d)", err);
		return 0;
	}

	LOG_INF("Scanning successfully started");

	for (;;) {

		if (default_conn) {
			if (k_sem_take(&data_received_sem, K_FOREVER)) {
				continue;
			}

			memset(data, atomic_get(&step), sizeof(data));

			err = bt_gatt_write_without_response_cb(
				default_conn, 21, data, sizeof(data), false, command_sent, NULL);
			if (!err) {
				pkt_cnt++;
			}
#if READ_RSSI_TEST
			if ((pkt_cnt % 10 == 0) && (!err)) {
				read_conn_rssi(default_conn->handle, &rssi);
			}
#endif // READ_RSSI_TEST
		}

		if (pkt_cnt % 100 == 0 && pkt_cnt != 0) {
			LOG_INF("100 packets sent.");
		}

		k_msleep(1);
	}
}
